#  Copyright 2022 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
from pathlib import Path
from typing import Any, Dict, Optional
from urllib.parse import urlparse

import yaml

from liga.io import open_uri
from liga.exceptions import SpecError
from liga.registry.base import ModelSpec, Registry, codegen_from_spec

__all__ = ["FileSystemRegistry"]


class FileModelSpec(ModelSpec):
    """File-based Model Spec.

    Parameters
    ----------
    raw_spec : dict
        The RAW dict passed from SparkSession
    validate : bool, default True.
        Validate the spec during construction. Default ``True``.

    Raises
    ------
    SpecError
        If the spec is not correct.
    """

    VERSION = "1.0"

    def __init__(
        self,
        raw_spec: Dict,
        need_validate: bool = True,
    ):
        self.base_dir: Optional[str] = None

        spec = {
            "name": raw_spec.get("name"),
            "options": raw_spec.get("options", {}),
            "model": {
                "plugin": raw_spec.get("plugin"),
                "type": raw_spec.get("modelType"),
                "uri": raw_spec.get("uri"),
            },
        }
        uri = spec["model"]["uri"]
        if not uri:
            raise SpecError("Model URI is missing")
        if Path(uri).suffix in [".yml", ".yaml"]:
            with open_uri(uri) as fobj:
                yaml_spec = yaml.load(fobj, Loader=yaml.FullLoader)
                spec = self._merge_spec(yaml_spec, spec)
                model_uri_from_yaml = yaml_spec.get("model", {}).get("uri")
                if model_uri_from_yaml is not None:
                    spec["model"]["uri"] = model_uri_from_yaml
            self.base_dir = os.path.dirname(uri)
        else:
            spec["version"] = self.VERSION

        super().__init__(spec, need_validate=need_validate)

    @staticmethod
    def _merge_spec(
        base_spec: Dict[str, Any], overwrite: Dict[str, Any]
    ) -> dict:
        """Merge `overwrite` spec into `base_spec`.

        All the values presented in `overwrite` will change the value in
        the `base_spec`.

        Parameters
        ----------
        base_spec : dict
            The base spec to be overwrite
        overwrite : dict
            The spec used to overwrite the base values
        """
        spec = base_spec.copy()
        for key, value in overwrite.items():
            if isinstance(value, dict):
                spec[key] = FileModelSpec._merge_spec(spec.get(key, {}), value)
            elif value is not None:
                spec[key] = value
        return spec

    def load_model(self):
        codegen = codegen_from_spec(self)
        return codegen.load_model_from_uri(self.model_uri)

    @property
    def model_uri(self):
        """Absolute model URI."""
        origin_uri = super().model_uri
        parsed = urlparse(origin_uri)
        if parsed.scheme or os.path.isabs(origin_uri):
            return origin_uri
        if self.base_dir is not None:
            return os.path.join(self.base_dir, origin_uri)
        return origin_uri


class FileSystemRegistry(Registry):
    """FileSystem-based Model Registry"""

    def __repr__(self):
        return "FileSystemRegistry"

    def make_model_spec(self, raw_spec: dict):
        return FileModelSpec(raw_spec)



#  Copyright 2020 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import re
import os
from importlib.metadata import version as find_version
from typing import Optional

from pyspark.sql import SparkSession

from liga.__version__ import version
from liga.logging import logger


def get_default_jar_version(use_snapshot: bool = True) -> str:
    """
    Make it easier to reference the jar version in notebooks and conftest.

    Parameters
    ----------
    use_snapshot: bool, default True
        If True then map `*dev0` versions to `-SNAPSHOT`
    """
    pattern = re.compile(r"([\d]+.[\d]+.[\d]+)")
    match = re.search(pattern, version)
    if not match:
        raise ValueError(f"Ill-formed version string {version}")
    match_str = match.group(1)
    if use_snapshot and (len(match_str) < len(version)):
        return match_str + "-SNAPSHOT"
    return match_str


def get_liga_assembly_jar(jar_type: str, scala_version: str) -> str:
    spark_version = find_version("pyspark").replace(".", "")
    name = f"liga-spark{spark_version}-assembly_{scala_version}"
    url = "https://github.com/komprenilo/liga/releases/download"
    github_jar = f"{url}/v{version}/{name}-{version}.jar"
    if jar_type == "github":
        if "dev" in version:
            logger.warning(
                "Jar type `github` is for stable release, "
                "it may fail when version contains dev"
            )
        return github_jar
    elif jar_type == "local":
        project_path = os.environ.get("ROOTDIR")
        if project_path:
            local_jar_path = f"{project_path}/out/liga/{scala_version}/assembly.dest/out.jar"
            if os.path.exists(local_jar_path):
                return local_jar_path
            else:
                raise ValueError("Please run `bin/mill 'liga[2.12].assembly'` first")
        else:
            logger.warning(
                "Jar type `local` is for developing purpose, fallback to Jar"
                " type `github` because no project root is specified"
            )
            return github_jar
    else:
        raise ValueError(f"Invalid jar_type ({jar_type})!")


def init_session(
    conf: Optional[dict] = None,
    app_name: str = "Liga",
    num_cores: int = 2,
    jar_type: str = "github",
    scala_version: str = "2.12",
) -> SparkSession:
    import sys

    os.environ["PYSPARK_PYTHON"] = sys.executable
    os.environ["PYSPARK_DRIVER_PYTHON"] = sys.executable

    default_conf = {
        "spark.sql.extensions": "net.xmacs.liga.spark.RikaiSparkSessionExtensions",
        "spark.driver.extraJavaOptions": "-Dio.netty.tryReflectionSetAccessible=true",
        "spark.executor.extraJavaOptions": "-Dio.netty.tryReflectionSetAccessible=true",
    }

    if conf and "spark.jars" in conf.keys():
        pass
    else:
        default_conf["spark.jars"] = get_liga_assembly_jar(
            jar_type, scala_version
        )

    if conf:
        for k, v in conf.items():
            default_conf[k] = v

    # Avoid reused session polluting configs
    active_session = SparkSession.getActiveSession()
    if active_session:
        for k, v in default_conf.items():
            if v is not None and str(active_session.conf.get(k)) != str(v):
                print(
                    f"active session: want {v} for {k}"
                    f" but got {active_session.conf.get(k)},"
                    f" will restart session"
                )
                active_session.stop()
                break

    builder = SparkSession.builder.appName(app_name)
    for k, v in default_conf.items():
        if v is not None:
            builder = builder.config(k, v)
            logger.info(f"setting {k} to {v}")
    session = builder.master(f"local[{num_cores}]").getOrCreate()
    return session


def init_dbr_session(app_name="Liga on Databricks"):
    return init_session(conf={"spark.jars": None}, app_name=app_name)



#  Copyright 2020 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Spark User Defined Types
"""

# Third Party
import numpy as np
from pyspark.sql import Row
from pyspark.sql.types import (
    ArrayType,
    BinaryType,
    IntegerType,
    StringType,
    StructField,
    StructType,
    UserDefinedType,
)

__all__ = [
    "NDArrayType",
]


class NDArrayType(UserDefinedType):
    """User define type for an arbitrary :py:class:`numpy.ndarray`.
    This UDT serializes :py:class:`numpy.ndarray` into bytes buffer.
    """

    def __repr__(self) -> str:
        return "np.ndarray"

    @classmethod
    def sqlType(cls) -> StructType:
        return StructType(
            fields=[
                # dtype field will use dictionary encoding.
                StructField(
                    "dtype",
                    StringType(),
                    False,
                ),
                StructField(
                    "shape",
                    ArrayType(IntegerType(), False),
                    False,
                ),
                StructField(
                    "data",
                    BinaryType(),
                    False,
                ),
            ]
        )

    @classmethod
    def module(cls) -> str:
        return "rikai.spark.types"

    @classmethod
    def scalaUDT(cls) -> str:
        return "org.apache.spark.sql.rikai.NDArrayType"

    def serialize(self, obj: np.ndarray):
        """Serialize an :py:class:`numpy.ndarray` into Spark Row"""
        return (
            str(obj.dtype),
            list(obj.shape),
            obj.tobytes(),
        )

    def deserialize(self, datum: Row) -> np.ndarray:
        import liga.numpy

        return (
            np.frombuffer(datum[2], dtype=np.dtype(datum[0]))
            .reshape(datum[1])
            .view(liga.numpy.ndarray)
        )

    def simpleString(self) -> str:
        return "ndarray"




# DO NOT EDIT THIS FILE -- AUTOGENERATED BY PANTS
# Target: //:liga

from setuptools import setup

setup(**{
    'install_requires': (
        'fsspec==2022.1.0',
        'jsonschema',
        'numpy',
        'pandas',
        'py4j',
        'pyarrow>=6.0',
        'pyspark==3.2.1',
        'pyyaml',
        'requests',
    ),
    'name': 'liga',
    'namespace_packages': (
    ),
    'package_data': {
    },
    'packages': (
        'liga',
        'liga.internal',
        'liga.numpy',
        'liga.registry',
        'liga.spark',
        'liga.spark.types',
    ),
    'python_requires': '<3.10,>=3.8',
    'version': '0.3.0.dev0',
})



#  Copyright 2022 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json
from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, Optional, TypeVar, List

from jsonschema.exceptions import ValidationError
from jsonschema.validators import validate

from liga.internal.reflection import find_func
from liga.exceptions import SpecError
from liga.logging import logger

__all__ = ["ModelSpec", "ModelType"]


M = TypeVar("M")  # Model Type


# JSON schema specification for the model payload specifications
# used to validate model spec input
def gen_schema_spec(required_cols: List) -> dict:
    return {
        "type": "object",
        "properties": {
            "version": {
                "type": "string",
                "description": "Model SPEC format version",
            },
            "schema": {"type": "string"},
            "name": {"type": "string", "description": "Model name"},
            "model": {
                "type": "object",
                "description": "model description",
                "properties": {
                    "uri": {"type": "string"},
                    "plugin": {"type": "string"},
                    "type": {"type": "string"},
                },
                "required": required_cols,
            },
        },
        "required": ["version", "model"],
    }


SPEC_PAYLOAD_SCHEMA = gen_schema_spec(["uri"])
NOURI_SPEC_SCHEMA = gen_schema_spec(["plugin", "type"])


def is_fully_qualified_name(name: str) -> bool:
    return "." in name


def parse_model_type(plugin: str, model_type: str) -> "ModelType":
    model_modules_candidates = []

    if is_fully_qualified_name(model_type):
        model_modules_candidates.append(model_type)
    elif is_fully_qualified_name(plugin):
        model_modules_candidates.extend(
            [
                f"{plugin}.models.{model_type}",
            ]
        )
    else:
        model_modules_candidates.extend(
            [
                f"liga.{plugin}.models.{model_type}",
            ]
        )
    for model_module in model_modules_candidates:
        try:
            return find_func(f"{model_module}.MODEL_TYPE")
        except ModuleNotFoundError:
            pass
    raise ModuleNotFoundError(
        f"Model type not found for plugin/model_type: {plugin}/{model_type}"
    )


class ModelSpec(ABC):
    """Model Spec Payload

    Parameters
    ----------
    spec : dict
        Dictionary representation of an input spec
    need_validate : bool, default True.
        Validate the spec during construction. Default ``True``.
    """

    def __init__(
        self,
        spec: Dict[str, Any],
        need_validate: bool = True,
        spec_schema: dict = SPEC_PAYLOAD_SCHEMA,
    ) -> None:
        self._spec = spec
        self._spec["options"] = self._spec.get("options", {})
        self._spec_schema = spec_schema
        if need_validate:
            self.validate()

    def validate(self) -> None:
        """Validate model spec

        Raises
        ------
        SpecError
            If the spec is not well-formatted.
        """
        logger.debug("Validating spec: %s", self._spec)
        try:
            validate(instance=self._spec, schema=self._spec_schema)
        except ValidationError as e:
            raise SpecError(e.message) from e
        if not self.plugin or not self.model_type:
            raise SpecError("Missing model plugin or model type")

    @property
    def version(self) -> str:
        """Returns spec version."""
        return str(self._spec["version"])

    @property
    def name(self) -> str:
        """Return model name"""
        return self._spec["name"]

    @property
    def model_uri(self) -> str:
        """Return Model artifact URI"""
        return self._spec["model"]["uri"]

    @property
    def model_type(self) -> "ModelType":
        """Return model type"""
        mtype = self._spec["model"].get("type", None)
        if mtype:
            return parse_model_type(self.plugin, mtype)
        else:
            raise SpecError("ModelType is not available in spec")

    @abstractmethod
    def load_model(self) -> Any:
        """Load the model artifact specified in this spec"""

    def load_label_fn(self) -> Optional[Callable]:
        """Load the function that maps label id to human-readable string"""
        if "labels" in self._spec:
            uri = self._spec["labels"].get("uri")
            if uri:
                with open(uri) as f:
                    data = json.load(f)
                return lambda label_id: data[label_id]
            func = self._spec["labels"].get("func")
            if func:
                return find_func(func)
        return None

    @property
    def plugin(self) -> str:
        """Model plugin"""
        return self._spec["model"].get("plugin", "")

    @property
    def schema(self) -> str:
        """Return the output schema of the model."""
        if "schema" in self._spec:
            return self._spec["schema"]
        return self.model_type.schema()

    @property
    def options(self) -> Dict[str, Any]:
        """Model options"""
        return self._spec["options"]


class ModelType(ABC):
    """Base-class for a Model Type.

    A Model Type defines the functionalities which is required to run
    an arbitrary ML models in SQL ML, including:

    - Result schema: :py:meth:`schema`.
    - :py:meth:`transform`, pre-processing routine.
    - :py:meth:`predict`, inference **AND** post-processing routine.
    """

    @abstractmethod
    def load_model(self, spec: ModelSpec, **kwargs: Any) -> None:
        """Lazy loading the model from a :class:`ModelSpec`."""
        pass

    @abstractmethod
    def schema(self) -> str:
        """Return the string value of model schema.

        Examples
        --------

        >>> model_type.schema()
        ... "array<struct<box:box2d, score:float, label_id:int>>"
        """
        pass

    @abstractmethod
    def transform(self) -> Callable:
        """A callable to pre-process the data before calling inference.

        It will be feed into :py:class:`torch.data.DataLoader` or
        :py:meth:`tensorflow.data.Dataset.map`.

        """
        pass

    @abstractmethod
    def predict(self, *args: Any, **kwargs: Any) -> Any:
        """Run model inference and convert return types into
        Rikai-compatible types.

        """
        pass

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return self.predict(*args, **kwargs)

    def release(self) -> None:
        """Release underneath resources if applicable.

        It will be called after a model runner finishes a partition in Spark.
        """
        # Noop by default
        pass



#  Copyright 2021 Rikai Authors
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import importlib
from abc import ABC, abstractmethod
from typing import Any, Tuple
from types import ModuleType, FunctionType

from pyspark.serializers import CloudPickleSerializer
from pyspark.sql.functions import udf
from pyspark.sql.types import BinaryType

from liga.internal.reflection import find_class
from liga.exceptions import SpecError
from liga.registry.model import ModelSpec, is_fully_qualified_name
from liga.logging import logger


__all__ = ["Registry"]


_pickler = CloudPickleSerializer()


class Registry(ABC):
    """Base class of a Model Registry"""

    @abstractmethod
    def make_model_spec(self, raw_spec: "ModelSpec") -> ModelSpec:
        """Make a ModelSpec from the raw model spec

        Parameters
        ----------
        spec : ModelSpec
        """

    def resolve(self, raw_spec: "ModelSpec") -> Tuple:
        """Resolve a model from the raw model spec.

        Parameters
        ----------
        spec : ModelSpec
        """
        name = raw_spec["name"]  # type: ignore[index]
        uri = raw_spec["uri"]  # type: ignore[index]
        if uri:
            logger.info("Resolving model %s from %s", name, uri)
        else:
            logger.info("Resolving model %s", name)
        return udf_from_spec(self.make_model_spec(raw_spec))

    def resolve_schema(self, raw_spec: "ModelSpec") -> str:
        """Resolve the model schema from the raw model spec."""
        name = raw_spec["name"]  # type: ignore[index]
        uri = raw_spec["uri"]  # type: ignore[index]
        if uri:
            logger.info("Resolving schema of model %s from %s", name, uri)
        else:
            logger.info("Resolving schema of model %s", name)
        spec = self.make_model_spec(raw_spec)
        return spec.schema


def codegen_from_spec(spec: ModelSpec) -> ModuleType:
    """Resolve the codegen module from the model spec.

    Parameters
    ----------
    spec : ModelSpec

    Returns
    -------
    ModuleType
        The imported module for the specific plugin codegen
    """
    if is_fully_qualified_name(spec.plugin):
        codegen_module = f"{spec.plugin}.codegen"
    else:
        codegen_module = f"liga.{spec.plugin}.codegen"

    try:
        codegen = importlib.import_module(codegen_module)
        return codegen
    except ModuleNotFoundError:
        logger.error("Unsupported model plugin: %s", spec.plugin)
        raise


def schema_from_spec(registry_class: str, raw_spec: dict) -> str:
    cls = find_class(registry_class)
    registry = cls()
    return registry.resolve_schema(raw_spec)


def udf_from_spec(spec: ModelSpec) -> Tuple:
    """Return a UDF from a given ModelSpec

    Parameters
    ----------
    spec : ModelSpec
       Model spec payload

    Returns
    -------
    udt_ser_func, udf_func, udt_deser_func, returnType
        Spark UDF function name for the generated data.
    """
    if spec.version != "1.0":
        raise SpecError(
            f"Only spec version 1.0 is supported, got {spec.version}"
        )

    def deserialize_return(data: bytes) -> Any:
        return _pickler.loads(data)

    codegen = codegen_from_spec(spec)
    return (
        pickle_udt,
        codegen.generate_udf(spec),  # type: ignore[attr-defined]
        deserialize_return,
    )


def command_from_spec(registry_class: str, raw_spec: dict) -> Tuple:
    cls = find_class(registry_class)
    registry = cls()
    return registry.resolve(raw_spec)


@udf(returnType=BinaryType())
def pickle_udt(data: Any) -> None:
    return _pickler.dumps(data)


def unpickle_transform(data: bytes) -> Any:
    return _pickler.loads(data)
